/*-----------------------------------------------------------------------------

   ROBOT.C

   CONTROL FUNCTIONS LIBRARY FOR RASCAL/LINUXBOT/MOOREBOT

   Copyright (c) 2000 Intelligent Autonomous Systems Laboratory
   University of the West of England, Bristol.  No I don't mean
   this - the intention is something along the lines of a GNU open
   source licence.

   author  Alan Winfield
   email   Alan.Winfield@uwe.ac.uk

   Note: for pretty layout set tabs=8
-----------------------------------------------------------------------------
 REVISION HISTORY:

 v1.1		15/9/00 -  About time to start some revision control, before it all
			   gets out of hand!
 v1.2		16/9/00 -  Fix bugs in getSwitches() and getSensors()
 v1.4		17/9/00 -  getAnalog() function added
 v1.5		10/10/00 - hardReset() changed to enable RASCAL ADC
 v1.6		8/4/01 -   proper calibration of timing loops added
 v1.7		8/4/01 -   getBattery() function added
 v1.8		11/4/01 -  bug in getAnalog() for Linuxbot fixed 
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <asm/io.h>				/* for inb() and outb() */
#include <sys/time.h>				/* for gettimeofday() */

#include "robot.h"

#define	DEBUG 1					/* set to one if debugging */


/*----------------------------------------------------------------------------
 Global variables start here...
 (these are variables not intended to be accessed by the robot programmer,
 as opposed to struct robot, which may be.)
-----------------------------------------------------------------------------*/

/* The number of busyLoop ticks per microsecond.. */
static int ticksperusec = 100;	/* default value in case programmer.. */
				/* forgets to call calibrate() */


/*----------------------------------------------------------------------------
 Actual code start here...
-----------------------------------------------------------------------------*/

/* Simple busy loop, for short delays of 10s of microseconds */
void busyLoop( int ticks )
{
	do {
		ticks--;
	} while (ticks);
}

/* Return the time, in microseconds, for a busyLoop() of 'ticks' loops.
   This will clearly only be accurate for large values of ticks, and
   even then only if the process is not interrupted. */
int timeLoop( int ticks )
{
	struct timeval starttime, stoptime;
	int tt;

	gettimeofday( &starttime, NULL );
	busyLoop( ticks );
	gettimeofday( &stoptime, NULL );
	tt = stoptime.tv_usec - starttime.tv_usec;
	if (tt < 0) tt+=1000000;	/* in case usecs have rolled over */
	return tt;
}

/* Return the number of busyLoop ticks per microsecond.
   (This may appear to be an awful lot of trouble for accurate usec
   timing loops, but udelay() is not available in user space. The
   alternative is to run the whole of robot.c under POSIX real-time
   scheduling, and use nanosleep() for delays; something to think about.)

   When compiled with optimisation on...
   a 100MHz 486 returns values of about 22
   a 450MHz PII returns about 220 
*/
int calibrate(void)
{
	int ticks, error;

	/* iterate in case timeLoop(100000) gets interrupted */
	do {
		ticks = 100000/timeLoop(100000);
		error = abs(100 - timeLoop( 100*ticks ));
	} while (error >= 10);	/* aim for error of less than 10% */
	return ticks;
}
		

/* Request Linux to allow access to the I/O ports.  Return non-zero value
   PORT_ACCESS_OK if access is granted, or 0 is access is denied.  Note:
   Access is only allowed from a process with root privileges (ie superuser). */
int accessPorts( void )
{
	if ( iopl( 3 ) < 0 )
		return 0;
	else
		return PORT_ACCESS_OK;
}


/* Write the value 'val' to the LEFT motion controller chip register 'reg'
   reg must have values between 0..63  */
void writeLeftMctrlReg(int reg, int val)
{
	outb(val, LEFT_MCTRL + (reg<<10));
	/* the delay in just in case of successive calls to this function,
	   which need to observe the minimum write cycle time = 1.83uS */
	busyLoop( 2 * ticksperusec );
}

/* Write the value 'val' to the RIGHT motion controller chip register 'reg'
   reg must have values between 0..63  */
void writeRightMctrlReg(int reg, int val)
{
	outb(val, RIGHT_MCTRL + (reg<<10));
	busyLoop( 2 * ticksperusec );
}


/* Return the value of the LEFT motion controller chip register 'reg'
   reg must have values between 0..63  */
int readLeftMctrlReg(int reg)
{
	(void) inb(LEFT_MCTRL + (reg<<10)); 	/* Two reads of slow HCTL 1100 */
	(void) inb(LEFT_MCTRL + (reg<<10));   	/* ..to setup read data */
	return (int) inb(LEFT_MCTRL2);        	/* Then read value from buffer */
}

/* Return the value of the RIGHT motion controller chip register 'reg'
	reg must have values between 0..63  */
int readRightMctrlReg(int reg)
{
	(void) inb(RIGHT_MCTRL + (reg<<10)); 	/* Two reads of slow HCTL 1100 */
	(void) inb(RIGHT_MCTRL + (reg<<10));   	/* ..to setup read data */
	return (int) inb(RIGHT_MCTRL2);        	/* Then read value from buffer */
}


/* Get the current value of the Read Actual Position registers for the
   Left motion controller.  */
long readLeftActualPosition(void)
{
	long	Pos  = readLeftMctrlReg(ACTUAL_POS_REG_LSB);
		Pos |= readLeftMctrlReg(ACTUAL_POS_REG)     << 8;
		Pos |= readLeftMctrlReg(ACTUAL_POS_REG_MSB) << 16;
	if (Pos & 0x800000) Pos |= 0xff000000;	/* 24 bit sign extend */
	return (Pos);
}


/* Get the current value of the Read Actual Position registers for the
   Right motion controller.  */
long readRightActualPosition(void)
{
	long	Pos  = readRightMctrlReg(ACTUAL_POS_REG_LSB);
		Pos |= readRightMctrlReg(ACTUAL_POS_REG)     << 8;
		Pos |= readRightMctrlReg(ACTUAL_POS_REG_MSB) << 16;
	if (Pos & 0x800000) Pos |= 0xff000000;	/* 24 bit sign extend */
	return (Pos);
}


/* Initialise the robot data structure pointed to by 'Bot' with all
   default and initial values.  This function must be called before any
   of the select...Mode functions.  */
void initBotData( robot *Bot )
{
	Bot->T = DEFAULT_T;
	Bot->K = DEFAULT_K;
	Bot->A = DEFAULT_A;
	Bot->B = DEFAULT_B;
	Bot->sampleperiod = (double)(DEF_SAMPLE_TIME);

	Bot->mode = UNDEFINED_MODE;

	Bot->maxmps = DEFAULT_MAXMPS;
	Bot->mpss   = DEFAULT_MPSS;
}


/* Perform a soft reset of both left and right motion controllers,
   and place then both in IDLE mode.  Note that in IDLE mode we can
   still move, in open loop.  */
void softReset( robot *Bot )
{
	/* Program Counter register.. */
	writeLeftMctrlReg(PC_REG, 0);    		/* software reset mode = 0 */
	writeLeftMctrlReg(PC_REG, 1);   		/* enter idle_mode = 1 */
	/* Flag register.. */
	writeLeftMctrlReg(FLAG_REG, F0_CLR);      	/* clear trapezoidal profile mode */
	writeLeftMctrlReg(FLAG_REG, F2_CLR);   		/* clear unipolar mode */
	writeLeftMctrlReg(FLAG_REG, F3_CLR);   		/* clear proportional velocity mode */
	writeLeftMctrlReg(FLAG_REG, F4_CLR);   		/* clear hold commutator flag */
	writeLeftMctrlReg(FLAG_REG, F5_CLR);   		/* clear integral velocity mode */

	/* Program Counter register.. */
	writeRightMctrlReg(PC_REG, 0);    		/* software reset mode = 0 */
	writeRightMctrlReg(PC_REG, 1);   		/* enter idle_mode = 1 */
	/* Flag register.. */
	writeRightMctrlReg(FLAG_REG, F0_CLR);      	/* clear trapezoidal profile mode */
	writeRightMctrlReg(FLAG_REG, F2_CLR);      	/* clear unipolar mode */
	writeRightMctrlReg(FLAG_REG, F3_CLR);   	/* clear proportional velocity mode */
	writeRightMctrlReg(FLAG_REG, F4_CLR);   	/* clear hold commutator flag */
	writeRightMctrlReg(FLAG_REG, F5_CLR);   	/* clear integral velocity mode */

	Bot->mode = IDLE_MODE;				/* set to idle mode */
}


/* Initialise the Left and Right motion controller digital filters to the
   values T,K,A and B in the Bot data structure.   */
void initFilter( robot *Bot )
{
	if (Bot->T < 16) Bot->T = 16;			/* see function below.. */
	writeLeftMctrlReg (SAMPLE_TIMER_REG, Bot->T);
	writeRightMctrlReg(SAMPLE_TIMER_REG, Bot->T);

	writeLeftMctrlReg (FILTER_GAIN_REG, Bot->K);
	writeRightMctrlReg(FILTER_GAIN_REG, Bot->K);

	writeLeftMctrlReg (FILTER_ZERO_REG, Bot->A);
	writeRightMctrlReg(FILTER_ZERO_REG, Bot->A);

	writeLeftMctrlReg (FILTER_POLE_REG, Bot->B);
	writeRightMctrlReg(FILTER_POLE_REG, Bot->B);

	/* finally recompute the sampleperiod based on supplied T value */
	Bot->sampleperiod = (16.0*(Bot->T+1)/CLOCK_FREQ);
}


/* The HCTL1100 manual states (last two paras of page 2-162) that new
   command positions and velocities should not be written to the device
   during the early part of each sample period.  In each sample period
   SAMPLE_TIMER_REG counts down from its initial value (set by initFilter,
   to zero).  The following two functions wait until SAMPLE_TIMER_REG is
   past this initial 15 increments.  Note that this means that in practice
   the minimum values of sample time are unusable.  */
void waitforLeftSafeTime( robot *Bot )
{	byte t;
	do {
		t = readLeftMctrlReg( SAMPLE_TIMER_REG );
	} while ( (t < Bot->T) && (t >= Bot->T-15) );
}

void waitforRightSafeTime( robot *Bot )
{	byte t;
	do {
		t = readRightMctrlReg( SAMPLE_TIMER_REG );
	} while ( (t < Bot->T) && (t >= Bot->T-15) );
}


/*  For both left and right motion controllers, initialise filter parameters,
    zero actual and command position registers, then select Position Control mode.  */
void selectPCMode( robot *Bot )
{
	softReset( Bot );				/* First go into IDLE */
	initFilter( Bot );                     		/* and initialise Filter */

	/* Now clear actual and command position registers.. */
	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (ACTUAL_POS_REG, 0);
	writeLeftMctrlReg (COMMAND_POS_REG_MSB, 0);
	writeLeftMctrlReg (COMMAND_POS_REG,     0);
	writeLeftMctrlReg (COMMAND_POS_REG_LSB, 0);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(ACTUAL_POS_REG, 0);
	writeRightMctrlReg(COMMAND_POS_REG_MSB, 0);
	writeRightMctrlReg(COMMAND_POS_REG,     0);
	writeRightMctrlReg(COMMAND_POS_REG_LSB, 0);

	// select position_control_mode
	writeLeftMctrlReg (PC_REG, 3);			/* just enter Control Mode = 3 */
	writeRightMctrlReg(PC_REG, 3);

	Bot->mode = PC_MODE;
}


/*  For both left and right motion controllers, initialise filter parameters,
    zero command velocity registers, then select Proportional Velocity mode.  */
void selectPVMode( robot *Bot )
{
	softReset( Bot );				/* First go into IDLE */
	initFilter( Bot );                     		/* and initialise Filter */

	/* Set initial velocity to 0.. */
	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (COMMAND_VEL_REG_MSB, 0);
	writeLeftMctrlReg (COMMAND_VEL_REG_LSB, 0);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(COMMAND_VEL_REG_MSB, 0);
	writeRightMctrlReg(COMMAND_VEL_REG_LSB, 0);

	/* select proportional velocity mode.. */
	writeLeftMctrlReg (PC_REG, 3);			/* first enter Control Mode */
	writeLeftMctrlReg (FLAG_REG, F3_SET);		/* then set Proportional velocity mode */
	writeRightMctrlReg(PC_REG, 3);			/* same again for right motor */
	writeRightMctrlReg(FLAG_REG, F3_SET);

	Bot->mode = PV_MODE;
}


/*  For both left and right motion controllers, initialise filter parameters,
    clear velocity and acceleration registers, then select Integral Velocity mode.  */
void selectIVMode( robot *Bot )
{
	softReset( Bot );				/* First go into IDLE */
	initFilter( Bot );                     		/* and initialise Filter */

	/* first select position control mode.. */
	writeLeftMctrlReg (PC_REG, 3);			/* just enter Control Mode = 3 */
	writeRightMctrlReg(PC_REG, 3);

	/* then set initial velocity and acceleration to 0.. */
	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (INT_COMMAND_VEL_REG, 0);
	writeLeftMctrlReg (ACCEL_REG_MSB, 0);
	writeLeftMctrlReg (ACCEL_REG_LSB, 0);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(INT_COMMAND_VEL_REG, 0);
	writeRightMctrlReg(ACCEL_REG_MSB, 0);
	writeRightMctrlReg(ACCEL_REG_LSB, 0);

	/* select integral velocity mode.. */
	writeLeftMctrlReg (FLAG_REG, F5_SET);
	writeRightMctrlReg(FLAG_REG, F5_SET);

	Bot->mode = IV_MODE;
}


/*  For both left and right motion controllers, initialise filter parameters,
    clear actual and command position registers, then select Trapezoid Profile mode.  */
void selectTPMode( robot *Bot )
{
	softReset( Bot );				/* First go into IDLE */
	initFilter( Bot );                     		/* and initialise Filter */

	/* Now clear actual and command position registers.. */
	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (ACTUAL_POS_REG, 0);
	writeLeftMctrlReg (COMMAND_POS_REG_MSB, 0);
	writeLeftMctrlReg (COMMAND_POS_REG,     0);
	writeLeftMctrlReg (COMMAND_POS_REG_LSB, 0);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(ACTUAL_POS_REG, 0);
	writeRightMctrlReg(COMMAND_POS_REG_MSB, 0);
	writeRightMctrlReg(COMMAND_POS_REG,     0);
	writeRightMctrlReg(COMMAND_POS_REG_LSB, 0);

	/* then select position control mode.. */
	writeLeftMctrlReg (PC_REG, 3);			/* just enter Control Mode = 3 */
	writeRightMctrlReg(PC_REG, 3);

	/* FLAG_REG will be set to F0_SET by the first command to move in this mode */
	Bot->mode = TP_MODE;
}


/* Select one of the five operating modes for the motion controllers,
   the default case is Idle (ie open loop) mode. */
void setMode( robot *Bot, int mode )
{
	switch( mode )
	{
	case PC_MODE:
		selectPCMode( Bot );
		break;
	case PV_MODE:
		selectPVMode( Bot );
		break;
	case IV_MODE:
		selectIVMode( Bot );
		break;
	case TP_MODE:
		selectTPMode( Bot );
		break;
	case IDLE_MODE:
	default:
		softReset( Bot );
	}
}


/* Open loop move with given demands of PWM duty cycle for left and right motors.
   In Idle mode we simply write to the 8-bit PWM motor port registers.  LeftCycle
   and RightCycle must both be in the range -100..100.  */
void openloopMove( robot *Bot, int LeftCycle, int RightCycle )
{
	/* We must be in Idle mode to be able to write to the PWM motor ports */
	if ( Bot->mode != IDLE_MODE )
		softReset( Bot );

	if (LeftCycle >  100) LeftCycle =  100;
	if (LeftCycle < -100) LeftCycle = -100;
	if (RightCycle >  100) RightCycle =  100;
	if (RightCycle < -100) RightCycle = -100;

	writeLeftMctrlReg (PWM_MOTOR_PORT, (byte) LeftCycle);
	writeRightMctrlReg(PWM_MOTOR_PORT, (byte) RightCycle);
}


/* Open loop stop.  This function works by forcing us into idle mode - whatever
   mode we're currently in - then setting the PWM motor ports to zero.  Note
   the robot will just coast to a halt, there will be no active braking.  */
void openStop( robot *Bot )
{
	openloopMove( Bot, 0, 0 );
}


/* Raw Closed loop move to a desired position.
   Both positions LeftP and RightP are signed 24-bit integers, which
   represent quadrature encoder counts.  */
void PCMove( robot *Bot, long LeftP, long RightP )
{
	if ( Bot->mode != PC_MODE )
		selectPCMode( Bot );

	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (COMMAND_POS_REG_MSB, ( LeftP  >> 16 ) & 0xff);
	writeLeftMctrlReg (COMMAND_POS_REG,     ( LeftP  >> 8  ) & 0xff);
	writeLeftMctrlReg (COMMAND_POS_REG_LSB, ( LeftP        ) & 0xff);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(COMMAND_POS_REG_MSB, ( RightP >> 16 ) & 0xff);
	writeRightMctrlReg(COMMAND_POS_REG,     ( RightP >> 8  ) & 0xff);
	writeRightMctrlReg(COMMAND_POS_REG_LSB, ( RightP       ) & 0xff);
}


/* Raw Closed loop move at a desired proportional velocity.
   Both velocities LeftV and RightV are signed 16-bit integers, which
   represent multiples of 16 quadrature encoder counts per sample time.
   Thus, we must prescale by 16, the desired velocities.  */
void PVMove( robot *Bot, long LeftV, long RightV )
{
	if ( Bot->mode != PV_MODE )
		selectPVMode( Bot );

	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (COMMAND_VEL_REG_MSB, ( LeftV  >>  8 ) & 0xff);
	writeLeftMctrlReg (COMMAND_VEL_REG_LSB, ( LeftV        ) & 0xff);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(COMMAND_VEL_REG_MSB, ( RightV >>  8 ) & 0xff);
	writeRightMctrlReg(COMMAND_VEL_REG_LSB, ( RightV       ) & 0xff);
}


/* Raw Closed loop move to a desired velocity, at a given acceleration.
   Velocities, LeftV and RightV are 8-bit signed integers. NOTE: successive
   calls should NOT have velocities differing by more than 127.

   Accelerations LeftA and RightA are unsigned 16-bit integers, which represent
   multiples of 256 quadrature encoder counts per sample time squared, so we
   must prescale by 256.  */
void IVMove( robot *Bot, long LeftV,  long LeftA,
			 long RightV, long RightA )
{
	if ( Bot->mode != IV_MODE )
		selectIVMode( Bot );

	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (ACCEL_REG_MSB,     ( LeftA >>  8 ) & 0xff);
	writeLeftMctrlReg (ACCEL_REG_LSB,     ( LeftA       ) & 0xff);
	writeLeftMctrlReg (INT_COMMAND_VEL_REG, LeftV & 0xff);
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(ACCEL_REG_MSB,     ( RightA >>  8 ) & 0xff);
	writeRightMctrlReg(ACCEL_REG_LSB,     ( RightA       ) & 0xff);
	writeRightMctrlReg(INT_COMMAND_VEL_REG, RightV & 0xff);
}


/* Raw Closed loop move to a given position, with a specified acceleration and a
   maximum velocity.

   Desired positions LeftP and RightP are signed 24-bit integers, which
   represent quadrature encoder counts.  Velocities LeftV and RightV are
   7-bit signed integers.  Accelerations LeftA and RightA are unsigned
   16-bit numbers which represent multiples of 256 quadrature encoder
   counts per sample time squared, so must be prescaled.  */
void TPMove( robot *Bot, long LeftP,  long LeftV,  long LeftA,
			 long RightP, long RightV, long RightA )
{
	if ( Bot->mode != TP_MODE )
		selectTPMode( Bot );

	waitforLeftSafeTime( Bot );
	writeLeftMctrlReg (ACCEL_REG_MSB,     ( LeftA >>  8 ) & 0xff );
	writeLeftMctrlReg (ACCEL_REG_LSB,     ( LeftA       ) & 0xff );
	writeLeftMctrlReg (MAX_VEL_REG,       ( LeftV       ) & 0x7f );
	writeLeftMctrlReg (FINAL_POS_REG_MSB, ( LeftP >> 16 ) & 0xff );
	writeLeftMctrlReg (FINAL_POS_REG,     ( LeftP >>  8 ) & 0xff );
	writeLeftMctrlReg (FINAL_POS_REG_LSB, ( LeftP       ) & 0xff );
	waitforRightSafeTime( Bot );
	writeRightMctrlReg(ACCEL_REG_MSB,     ( RightA >>  8 ) & 0xff );
	writeRightMctrlReg(ACCEL_REG_LSB,     ( RightA       ) & 0xff );
	writeRightMctrlReg(MAX_VEL_REG,       ( RightV       ) & 0x7f );
	writeRightMctrlReg(FINAL_POS_REG_MSB, ( RightP >> 16 ) & 0xff );
	writeRightMctrlReg(FINAL_POS_REG,     ( RightP >>  8 ) & 0xff );
	writeRightMctrlReg(FINAL_POS_REG_LSB, ( RightP       ) & 0xff );

	/* Now set the flag to start the trapezoid profile move
	   The flag clears after the last internal move command has been issued.
	   This is likely to be just before it finishes.  */
	writeLeftMctrlReg (FLAG_REG, F0_SET);
	writeRightMctrlReg(FLAG_REG, F0_SET);
}


/* Wait for a trapezoidal move to complete.  This function MUST be called
   before the 2nd or subsequent trapezoidal profile moves. */
void waitforTPMove( robot *Bot )
{
	unsigned char lpflag, rpflag;
	do {
		/* read and mask the left and right trapezoidal profile flags.. */
		lpflag = readLeftMctrlReg ( STATUS_REG ) & 0x10;
		rpflag = readRightMctrlReg( STATUS_REG ) & 0x10;
	} while (lpflag | rpflag);   	/* loop until BOTH are clear */
}


/* Convert the real number velocity mps, in metres per second, to
   quadrature encoder counts, for the curremt sample time value. */
long convertMPStoCount( robot *Bot, float mps )
{
	long v = (long)(COUNTS_PER_METRE * mps * Bot->sampleperiod);
	return v;
}


/* Convert the real number acceleration mpss, in metres per second per second,
   to quadrature encoder counts, for the current sample time value. */
long convertMPSStoCount( robot *Bot, float mpss )
{
	double s = Bot->sampleperiod;
	long a = (long)(256.0 * COUNTS_PER_METRE * mpss * s * s);
	if ( a < 0 ) a = 0;	/* check.. -ve accelerations? */
	else if ( a > 0x7fff ) a = 0x7fff;
	return a;
}


/* Execute a closed-loop move, by real-number distances LeftD and RightD metres,
   for the left and right wheels respectively.

   This function uses either position control mode, or trapezoidal profile mode
   and assumes that one of these modes has already been selected.  If neither
   of these modes has been selected, then do not attempt to move, and return
   with the non-zero value WRONG_MODE_ERROR.  Otherwise return 0 (success).
*/
int posMove( robot *Bot, float LeftM, float RightM )
{
	long LeftP, RightP, maxvel, accel;

	/* Read current shaft encoder values, and add the demand offsets for
	   this move.. */
	LeftP  = readLeftActualPosition()  + (long)(LeftM * COUNTS_PER_METRE);
	RightP = readRightActualPosition() + (long)(RightM * COUNTS_PER_METRE);

	switch ( Bot->mode )
	{
	case PC_MODE:
		PCMove( Bot, LeftP, RightP );
		break;

	case TP_MODE:
		maxvel = convertMPStoCount( Bot, Bot->maxmps );
		accel = convertMPSStoCount( Bot, Bot->mpss );
		TPMove( Bot, LeftP,  maxvel, accel,
			     RightP, maxvel, accel );
		break;

	 default:
		return WRONG_MODE_ERROR;
	}
	return 0;
}


/* Execute a Rotation on-the-spot, by Rot degrees.  Rot MUST be within the
   range -360.0 .. +360.0.  This function assumes we are already in one of
   the two closed loop position control modes.

   Return 0 on success, or WRONG_MODE_ERROR if we're in the wrong mode.
*/
int rotate( robot *Bot, float Rot )
{
	float x = Rot*PI*WHEEL_BASE/360.0;
	/* rotate clockwise for +ve angles.. */
	return posMove( Bot, x, -x );
}


/* Execute a move at given real number velocities, in m/s, for the left and
   right wheels respectively.

   This function uses either one of the two closed loop velocity modes:
   Proportional velocity mode or Integral velocity mode, or if we are in
   idle mode, then use (the highly innacurate) open loop PWM control move.

   If the robot is in any other mode, then do not attempt a move, and return
   with the non-zero WRONG_MODE_ERROR.  Otherwise return 0 (success).
*/
int velMove( robot *Bot, float LeftV, float RightV )
{
	long lv, rv, accel;

	switch ( Bot->mode )
	{
	case PV_MODE:
		lv = 16 * convertMPStoCount( Bot, LeftV );
		rv = 16 * convertMPStoCount( Bot, RightV );
		PVMove( Bot, lv, rv );
		break;
	case IV_MODE:
		lv = convertMPStoCount( Bot, LeftV );
		rv = convertMPStoCount( Bot, RightV );
		accel = convertMPSStoCount( Bot, Bot->mpss );
		IVMove( Bot, lv, accel, rv, accel );
		break;
	case IDLE_MODE:
		/* note: highly inaccurate conversion from mps to PWM cycles */
		openloopMove( Bot, (int)(100.0*LeftV), (int)(100.0*RightV) );
		break;
	default:
		return WRONG_MODE_ERROR;
	}
	return 0;
}


/*  Perform a hard reset of the Motion Controller, and enable the motors.
    This function needs to be seriously improved for proper timing and to
    remove magic numbers */
void hardReset( void )
{
#ifdef LINUXBOT
	outb( 0, OPLATCH1 );  		/* ..also clears LEDs */
	busyLoop( 50 * ticksperusec );	/* 50uS reset pulse is plenty */
	outb( 3, OPLATCH1 );
#else	/* RASCAL */
	outb( 0x17, ENABLE_PORT );
	busyLoop( 50 * ticksperusec );
	outb( 0x1f, ENABLE_PORT );
#endif
}

/*------------------------------------------------------------------------------
  Code not just concerned with motion control starts here...
------------------------------------------------------------------------------*/

/* Set a value to the robot's LEDs,
   for the RASCAL there are 8 LEDS, so this can be a value 0..0xff,
   for the LinuxBot/MooreBot there are just 5 LEDS, so values 0..0x1f */
void setLEDs( int value )
{
#ifdef LINUXBOT
	/* the OR 3 is in order to preserve the value written by hardReset() */
	outb( ((value & 0x1f )<<3) | 3, OPLATCH1 );
#else	/* RASCAL */
	outb( value & 0xff, LED_PORT );
#endif
}


/* Get the value of the robot's bump sensors */
int getSensors( void )
{
#ifdef LINUXBOT
	/* on the LinuxBot the bump sensors are bits 4..7 of IP Latch 1 */
	return ( (inb(IPLATCH1) >> 4) & 0x0f );
#else	/* RASCAL */
	/* on the RASCAL the bump sensors are bits 2..3 of the status port */
	return ( (inb(STATUS_PORT) & 0x0c) >> 2 );
#endif
}


/* Get the value of the robot's status switches */
int getSwitches( void )
{
#ifdef LINUXBOT
	/* on the LinuxBot the status switches are bits 0..3 of IP Latch 1 */
	return ( inb(IPLATCH1) & 0x0f );
#else	/* RASCAL */
	/* on the RASCAL the status switches are bits 0..1 of the status port */
	return ( inb(STATUS_PORT) & 0x03 );
#endif
}


/* Get analog (A/D) inputs...
   For the LinuxBot/MooreBot there are 8 12-bit A/D channels (0..7).
   For the RASCAL there are 4 8-bit A/D channels (0..3).
*/
float getAnalog( int channel )
{
	byte lowbyte, highbyte;

#ifdef LINUXBOT
	if (channel < 0 || channel > 7)  	/* check channel in range */
		return 0;
	outb( channel+0x50, ADC_WRITE );	/* set channel and start */
	busyLoop( 20 * ticksperusec  );		/* wait 20uS for conversion */
	lowbyte  = inb( ADC_READ_LOW );		/* read low byte.. */
	highbyte = inb( ADC_READ_HIGH );	/* ..and high 4 bits */

	/* The MAX199 A/D converter has an internal 4.096V reference,
	   so we just divide the 12-bit result by 1000 */
	return (highbyte<<8 | lowbyte) / 1000.0;

#else   /* RASCAL */
	/* Here we assume the voltage reference has already been enabled.. */
	if (channel < 0 || channel > 3)  	/* check channel in range */
		return 0;
	outb( channel+4, ADC_WRITE );		/* set channel and start */
	busyLoop( 40 * ticksperusec );		/* wait 40uS for conversion */
	lowbyte  = inb( ADC_READ );		/* read result */

	/* The RASCAL A/D converter uses an external 2.5V reference,
	   so 1 bit equals 2.5/256 volts */
	return (lowbyte * 2.5 / 256.0);
#endif
}


/* Get the current battery voltage.
   This function assumes the proper jumper is in place to connect the
   battery voltage to channel zero of the A/D converter.
*/
float getBattery( void )
{
	return ( getAnalog(0) * BATT_SCALE );
};


/* Full Reset of the Robot */
void resetBot( robot *Bot )
{
	ticksperusec = calibrate();	/* 1st calibrate timing loops */
#if DEBUG
	printf("Ticks per microsecond = %d\n", ticksperusec);
#endif
	hardReset();		/* First hard reset, and enable motors */
	initBotData( Bot );	/* then initialise robot data structure */
	softReset( Bot );	/* soft reset, and go into idle mode */
	initFilter( Bot );	/* initialise motion controller filter params */
	openStop( Bot );	/* make sure we're stopped */
	setLEDs( 0 );           /* and finally clear the LEDs */
	getBattery();		/* dummy 1st read of A/D (since the 1st.. */
				/* reading tends to be duff */
}


/* example code:

robot Bot;

if ( !accessPorts() )
{
	printf("Port access denied.. root privileges needed\n");
	exit(1);
}

resetBot( &Bot );
setMode( &Bot, TP_MODE );
posMove( &Bot, 0.5, 0.5 );
waitforTPMove( &Bot );
rotate( &Bot, 90.0 );
setMode( &Bot, IV_MODE );
velMove( &Bot, 0.8, 0.8 );


*/




